open Ast

(* TITLE:	THE PRECOMPILER *)
(* AUTHOR:	PETER BURROWS	*)
(* DATE:	7/2/15			*)


(*=================  THE ENVIRONMENT  =================*) 

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

let binmap = NameMap.empty;;

let objects = NameMap.empty;;



(*			
	----------------------------------------------- 	
	NOTE: binmap takes the following module form -> 
			(string [variables] * module [objects])

	NOTE: objects take the following tuple form -> 
			(string [object], (int [value] * int array [bit indices]) 

	----------------------------------------------- *)

(* Creating/Updating the environment: *)

(* Initialize mapUpdate:
	- Note: this is a place holder for other global functions to initialize
	- mapUpdate is defined later *)
let mapUpdate = (fun var obj tuplVal map -> 
		if (NameMap.mem var map) then
			let objects = NameMap.find var map in
			let objects = NameMap.add obj tuplVal objects in
			let map = NameMap.add var objects map in map
		else
			let objects = NameMap.add obj tuplVal objects in
			let map = NameMap.add var objects map in map);;



(* Subfunctions to retrieve the value and bit indices of an object *)
let gettuple = (fun var obj map -> NameMap.find obj (NameMap.find var map) );;
let getintVal = (fun (w,x,y,z) -> w);;
let getBitIndices = (fun (w,x,y,z) -> x);;
let getBinlist = (fun (w,x,y,z) -> y);;
let getSpecialvar = (fun (w,x,y,z) -> z);;




(* ------------------------------updateBinlist --------------------------------*)
(* Update object 'a' given the new value of new object 'newb' and output the new 'a' *)
let getInd = (fun (x,y) -> x);;
let getBit = (fun (x,y) -> y);;

let rec comp a b inc = 
	match a with
	  [] -> []
	| hd::tl -> 
		if ( (getInd (hd)) = (getInd (List.nth b inc) ) ) then
			(List.nth b inc) :: comp tl b (inc) else (hd)::comp tl b (inc);; 

let rec compall a b inc =
	if inc > -1 then compall (comp a b inc) b (inc-1) else a;;

let updateBinlist a newb =
	compall a newb ( (List.length newb) - 1);;


(* ---------------------------lists2tuplelist --------------------------------*)
(* Merge the binary indices list and binary list to a combined tuple list 
	i.e. [1;2;3] [4;5;6] -> [(1,4);(2,5);(3,6)] *)
	(* NOTE: l1 and l2 must be the same length *)
let rec lists2tuplelist l1 l2 = 
	match l1 with
	  [] -> []
	| hd::tl -> (hd,List.hd l2):: lists2tuplelist tl (List.tl l2);;


(* -----------------------------tuplelist2list --------------------------------*)
(* Convert the tuple lists to lists *)
let rec tuplelist2list l ind = 
	match l with
	  [] -> []
	| hd::tl -> if ind = 0 then (fun (x,y) -> x) hd :: tuplelist2list tl ind
				else if ind = 1 then (fun (x,y) -> y) hd :: tuplelist2list tl ind
				else [];;


(* ============================= TYPE CONVERSIONS =================================*)
(* ===== binstr2binlist: Convert binary string to binary int list:

	i.e. "110010" -> [1;1;0;0;1;0]

														*)
let binstr2binlist s =
	let str2charlist s =
  		let rec exp i l =
    		if i < 0 then l else exp (i - 1) (s.[i] :: l) in
  		exp (String.length s - 1) [] in
	List.map (fun a -> (int_of_char a) - 48) (str2charlist s);;

(* ===== binlist2int : Convert binary int list to integer:
	
	i.e. [1;0;0;0] -> 8 
														*)
let binlist2int binlist =
  let converted_binlist = List.map float_of_int binlist in
  let helper converted_binlist = 
	let rec eval i = function
	   []	 -> []
  	| hd::tl -> if i < 0. then tl else (hd *. 2. ** i) :: eval (i -. 1.) tl in
	eval (float_of_int(List.length converted_binlist - 1)) converted_binlist in
  int_of_float (List.fold_left (fun s e -> s +. e) 0. (helper converted_binlist) );;


(* ===== int2binlist: Convert int to a binary list 

	i.e. 18 -> [1;0;0;1;0]
														*)
let int2binlist intVal = 
let intVal = float_of_int intVal in
  let rec eval l intVal =
	let quot = (intVal) /. 2. in
	let flquot = floor(quot) in
	let r = 2. *. (quot -. flquot) in
	  if quot < 1. then r::l else r::eval l flquot in 
  let l = eval [] intVal in 
let l = List.rev (l) in
List.map int_of_float l;;
(* ===================== END OF TYPE CONVERSIONS ===================================*)



(* =========== CONSTRUCTED SHIFT OPERATORS *)
(* ===== shiftR: i.e shiftR [1;1;1] 1 -> [0;1;1] 	*)
let shiftR l byX = 
let rec helper fill l byX =
	match l with
	  [] -> fill
	| hd::tl ->
		if (byX > 0) then (
			helper (0::fill) tl (byX-1)
		) else 
			fill@(List.rev(l)) in
helper [] (List.rev(l)) byX;;

(* ===== shiftL: i.e shiftL [1;1;1] 1 -> [1;1;0] 	*)
let shiftL l byX = 
let rec helper fill l byX =
	match l with
	  [] -> fill
	| hd::tl ->
		if (byX > 0) then (
			helper (0::fill) tl (byX-1)
		) else 
			(l)@fill in
helper [] (l) byX;;




(* ------------------------ mapUpdate ----------------------------*)
exception Failure of string;;


(* ===== updater : 
	- the final subfunction executed by mapUpdate 
	- updater updates the values of all objects within a root variable,
		given a chage in the root variable, or one of the objects *)

	(* ===== gettuplelist - retrieves the indices and binary list from the environment *)
	let gettuplelist var obj map =
	  lists2tuplelist 
		(getBitIndices (gettuple var obj map)) (getBinlist (gettuple var obj map));;

(* ===== updaterhelper: subfunction called within updater *)
let rec updaterhelper var newObj objNames map = 
	match objNames with
	  [] -> map
	| hd::tl -> let old = gettuplelist var hd map in
				let old = updateBinlist old newObj in
		  let map = mapUpdate var hd
			( binlist2int (tuplelist2list old 1),
	   		  (tuplelist2list old 0),
		 		(tuplelist2list old 1),
		   		  (getSpecialvar(gettuple var hd map)) ) 
		  map in updaterhelper var newObj tl map;;

(* ===== updater *)
let updater var newObj map = 
	updaterhelper var newObj (List.rev(NameMap.fold (fun s c out -> s::out) 
							(NameMap.find var map) []) ) map;;



(* ===== check: called within mapUpdateHelper1 (which is declared below)
	- throw failures when new objects exceed previously declared bounds/indices *)
let check var obj tuplVal map =

  (* general check to make sure that the objects do not use duplicate indices *)
  let rec duplicates l = 
  match l with
    [] -> []
  | hd::tl -> hd::(duplicates (List.filter (fun x -> x <> hd) tl)) in
  let out = (duplicates(getBitIndices(tuplVal))) in
	if ( (List.length out) < (List.length (getBitIndices(tuplVal))) ) then ( 
		raise (Failure ((var)^"."^(obj)^" uses duplicate indices") )
    ) else [];

  (* general check to make sure that the objects have identical indices *)
  if ( (obj <> "") && (NameMap.mem var map) ) then (

		(* compare tupl lengths *)
		if ( (List.length (getBitIndices(tuplVal))) <  
			 (List.length (getBinlist(tuplVal))) ) then (
		    raise (Failure ((var)^"."^(obj)^" overflow"))
		) else [];

		if (NameMap.mem obj (NameMap.find var map)) then (
			if ( (getBitIndices(gettuple var obj map)) <> 
				 (getBitIndices(tuplVal)) ) then ( 
				raise (Failure ((var)^"."^(obj)^" changed its previous bounds")) 
			) else []
		) else []
  ) else [] ;

  (* if the variable has already been declared and if it is the root *)
  if ( (NameMap.mem var map) && (obj = "") ) then ( 

		(* if the root has already been declared,
			make sure the new declaration does not exceed its previous value.. *)
		if ((NameMap.mem "" (NameMap.find var map)) &&
		( (List.nth (getBitIndices(gettuple var "" map)) 0) < 
		  (List.nth (getBitIndices(tuplVal)) 0)) ) then ( 
			raise (Failure ((var)^" exceeds its previous bounds"))

		(* ...else if the new root has less indices than its previous value.. *)
		) else if ((NameMap.mem "" (NameMap.find var map)) &&
		( (List.nth (getBitIndices(gettuple var "" map)) 0) > 
		  (List.nth (getBitIndices(tuplVal)) 0)) ) then (
			(* do nothing -> the root indices will be reformatted in mapUpdate's 1st check *)
			[]

		(* ..else compare new root to existing objects (if any) *)
		) else ( 
		let rec outer objname =
		match objname with 
		  [] -> []
		| hd::tl ->
		  (* compare root to all of the indices of all of the objects *)
		  let rec compareObjs ind = 
			  if ((ind > -1) && ( (List.nth (getBitIndices(tuplVal)) 0) < 
				(List.nth (getBitIndices(gettuple var hd map)) ind) ) ) then (
					raise (Failure ((hd)^" exceeds the bounds of "^(var)))
			  ) else if (ind > -1) then ( compareObjs (ind-1) 
			  ) else outer tl in
		  compareObjs ((List.length (getBitIndices (gettuple var hd map)))-1) in
		outer (List.rev(NameMap.fold (fun s c out -> s::out) (NameMap.find var map) [])) )

  (* if the variable has already been declared and the object is not the root,
	 but the root does exist *)
  ) else if ( (NameMap.mem var map) && ( NameMap.mem "" (NameMap.find var map)) ) then (
		let rec compareObj2root newobjIndices  =
		match newobjIndices with
		  [] -> []
		| hd::tl -> if ( hd > (List.nth (getBitIndices(gettuple var "" map)) 0) ) then (
						raise(   Failure ((obj)^" exceeds the bounds of "^(var)))
					) else compareObj2root tl in
		compareObj2root (getBitIndices(tuplVal))

  (* if the variable has not been declared *)
  ) else [];;
	


(* ===== mapUpdateHelper 1: called within mapUpdateHelper0, which is seen below *)
let rec mapUpdateHelper1 = (fun var obj tuplVal map -> 
  (* Check the bounds of the new object *)
  let _ = check var obj tuplVal map in
  
  if (NameMap.mem var map) then (
	let objects = NameMap.find var map in
	(* If objects exists *)
	if (NameMap.mem obj objects) then (

	  (* diff MISMATCH procedure left pads zeros to maintain the binary array, as needed *)
	  let diff = ((List.length (getBinlist(NameMap.find obj objects)) ) -
		     (List.length (getBinlist(tuplVal)))) in
	    if (diff < 0) then
			raise (Failure ((var)^"."^(obj)^" length mismatch/overflow") )
	    else if (diff > 0) then ( (* concat based on diff and add indices *)
		  let rec append l n =
			if n = 0 then l else 0::append l (n-1) in 
			let appended = append (getBinlist(tuplVal)) diff in
			let objects = NameMap.add obj 
				( binlist2int(appended),
					(getBitIndices(NameMap.find obj objects)),
					 (appended),
						(getSpecialvar(NameMap.find obj objects)) ) objects in
			let map = NameMap.add var objects map in map)
		 (* populate the map with the modified object *)
	     else
			let objects = NameMap.add obj tuplVal objects in
			let map = NameMap.add var objects map in map)

	(* else load a new variable and object/root *)
	else (
		let objects = NameMap.add obj tuplVal objects in
		let map = NameMap.add var objects map in map)) 

  (* else load a new variable and object/root *)
  else(	
	let objects = NameMap.add obj tuplVal objects in
	let map = NameMap.add var objects map in map));;


(* ===== mapUpdateHelper0: first subfunction called by mapUpdate *)
let mapUpdateHelper0 var obj tuplVal map = 
  (* First Check: 
	  - reformat the root variable if its new indices are smaller
	      than its previous indices *)
  if ( (obj ="") && (NameMap.mem var map) && (NameMap.mem "" (NameMap.find var map)) &&
	 ( (List.nth (getBitIndices(gettuple var "" map)) 0) >= 
	   (List.nth (getBitIndices(tuplVal)) 0)) ) then (

		(* compare new root indices with the old *)
		let rec compareRootInds newRoot ind = 
		match newRoot with
		  [] -> []
		| hd::tl -> if (hd = (List.nth 
						(List.rev(getBitIndices(gettuple var "" map))) ind)) then (
						compareRootInds tl (ind+1)
					) else (
						raise (Failure ((var)^"'s indices have changed")) ) in 
		compareRootInds (List.rev (getBitIndices(tuplVal))) 0; 
		
			(* then update the map using prior root indices *)
			(mapUpdateHelper1 var obj 
				( getintVal(tuplVal),
					getBitIndices(gettuple var "" map),
						getBinlist(tuplVal),
							getSpecialvar(tuplVal) ) map) 
	) else mapUpdateHelper1 var obj tuplVal map;;



(* ===== mapUpdate *)
let mapUpdate var obj tuplVal map =
	let map = mapUpdateHelper0 var obj tuplVal map in
	let map = updater var (gettuplelist var obj map) map in
	map;;
(* --------------------END OF mapUpdate ----------------------------*)




(* ================== Additional global functions ====================== *)

(* ===== mkIndices -> makes an int list given a length
	i.e. mkIndices [4;2;3] (List.length [4;2;3]) 
			-> [2;1;0]] *)
let rec mkIndices l len =
		if (len > -1) then (
			len::mkIndices l (len-1)
		) else l;;

(* print_list -> used for debugging *)
let rec print_list = function 
			  [] -> print_endline "]"
			| e::l -> print_int e ; 
					if (List.length l = 0) then print_string ""
					else print_string ";" ; print_list l;;

(* print_string_list -> used for debugging *)
let rec print_string_list = function 
			  [] -> print_endline "]"
			| e::l -> print_string e ; 
					if (List.length l = 0) then print_string ""
					else print_string ";" ; print_string_list l;;

(* checktotaloperations ->
	Check length of the IR vhdl string list. If too many assignment operations 
	(i.e. var := a+b+c), then raise an error *)
let checktotaloperations = (fun appendvhdl ->
  let checktotaloperations =
	if ( (List.length (appendvhdl)) > 8 ) then (
	  let rev = List.rev (appendvhdl) in
	  let var = List.hd (rev) in
	  let obj = List.hd (List.tl rev) in
        raise (Failure ("Too many arithmetic operators assigned to "^(var)^"."^(obj)) )			
		  ) else [] in checktotaloperations);;


(* print -> i.e. print var obj -> prints to screen *)
	let print var obj binmap = [];
		print_endline "\n";
	  	print_string var; print_string "."; print_string obj; print_endline " =";
      	print_string "\tInteger Representation:\t\t"; 
			print_endline (string_of_int (getintVal(gettuple var obj binmap)));
	  	print_string "\tBinary Indices:\t\t\t";
			print_string "[";
			print_list (getBitIndices(gettuple var obj binmap));
	  	print_string "\tBinary Representation:\t\t";
			print_string "[";
			print_list (getBinlist(gettuple var obj binmap));
	  	(*print_string "\tType (In/Out/Signal):\t\t";
			print_endline (string_of_int (getSpecialvar(gettuple var obj binmap)));*)
	  	print_endline "\n";;

(* ================== END OF Additional global functions ======================*)



(* INITIALIZE THE VHDL IR FOR EVALUATION *)
print_string "\n======================== blooRTLs ===========================>";;
let vhdl = NameMap.empty;;
let vhdlcount = "0";;
let vhdl = NameMap.add vhdlcount (""::[]) vhdl;;

(* vhdl incrementer *)
let increment =(function count -> string_of_int((int_of_string count) + 1));;

(* -----------------------------------------------------------------------*)
(* ---------------------------- EVALUATION ------------------------------ *)
(* -----------------------------------------------------------------------*)
(* ==== evaluates BINMAP statements *)
let rec evalBinmap var obj indices binmap = (function
	Objdeclseq(obj,decl1,decl2) -> ( (*print_string "\nObjdeclseq(obj,decl1,decl2) ->";*)
	  let var, obj, indices, binmap = evalBinmap var obj indices binmap decl1 in 
	  let var, obj, indices, binmap = evalBinmap var "" [] binmap decl2 in
	  var, obj, indices, binmap )
 
  | Objmap(obj,decl) -> ( (*print_string "\nObjmap(obj,decl) -> ";*)
	  let var, obj, indices, binmap = evalBinmap var obj indices binmap decl in
	  var, obj, indices, binmap )

  |	Indseq(ind,decl) -> ( (*print_string "\n\tIndexSeq(i,decl) -> ";*)
	  let var, obj, indices, binmap = evalBinmap var obj (ind::indices) binmap decl in
	  var, obj, indices, binmap )
		
  |	Indices(ind) -> ( (* print_string "Index(i) -> "; *)
	let ind = ind::indices in
	  let dummybits = List.map (fun x -> x-x) ind in
	  let tupl = (0,List.rev(ind),dummybits,0) in

	  (* if the variable already exits... *)
	  if ((NameMap.mem var binmap) && 
			(NameMap.mem "" ((NameMap.find var binmap)))) then (
		(* retrieve all values from the root variable... *)
		let originaltupl = gettuple var "" binmap in
		  (* then, update the new object indices to the map... *)
		  let binmap = mapUpdate var obj tupl binmap in 
			(* finally, restore original values of the root to the map *)
			let binmap = mapUpdate var "" originaltupl binmap in
	(* return: *)	
	 var, obj, indices, binmap

	  (* otherwise, initialize the object loading the dummy bit values *)
	  ) else (
		let binmap = mapUpdate var obj tupl binmap in

	(* return: *)	
	  var, obj, indices, binmap ) )

)

(* ==== evaluates Arithmetic Expressions/Assignments *)
let rec eval var obj binmap vhdl vhdlcount = (function
(* all types return (tuplVal, map) *)
	
	Bits(b) -> ( (* print_string "LoadBits(b) -> ";*)
	  let binlist = binstr2binlist b in
		let x = binlist2int binlist in
		  let bitIndices = mkIndices [] ((String.length b)-1) in

			(* APPEND VHDL *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = b::"BITS"::appendvhdl in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

		var, obj, vhdl,vhdlcount,(x,bitIndices,binlist,-1), binmap )

  | Lit(x) -> (		(* print_string "LoadInt(i) -> ";*)

	(* convert ['0'- '9']+'d' to integer *)
	let intx = 
		let removeD s =
			let rec helper s i out =
	  			if ( i < 0 ) then (
		 		  out
	 			) else  ( helper s (i-1) (out + ( (int_of_char s.[i]) - 48)*
				  int_of_float(10.**(float_of_int(((String.length s) - 2) - i)))) ) in
			helper s (((String.length s) - 2)) 0 in 
		removeD x in
	

		(* APPEND VHDL *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = (string_of_int intx)::"INT"::appendvhdl in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

	    (* return the tupl and binmap *)
		var, obj, vhdl,vhdlcount,(intx,[],(int2binlist intx),-1), binmap)

  | IdenRoot(var) -> (	(*print_string "IdenRoot(var) -> "; *)

	if ( (NameMap.mem var binmap) && (NameMap.mem "" (NameMap.find var binmap)) ) then (
		let newintVal = getintVal (gettuple var "" binmap) in
		let newBitIndices = getBitIndices (gettuple var "" binmap) in
		let newBinlist = getBinlist (gettuple var "" binmap) in
		let newSpecialVar = getSpecialvar (gettuple var "" binmap) in

			(* APPEND VHDL *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = ""::var::appendvhdl in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

	var, obj, vhdl,vhdlcount, (newintVal,newBitIndices,newBinlist,newSpecialVar), binmap 
	(* if iden not found, then raise an error! *)
	) else (raise (Failure ((var)^" has no value")))  )

  | IdenObj(var, obj) -> (	(*print_string "IdenObj(var,obj) -> "; *)

	if ( (NameMap.mem var binmap) && (NameMap.mem obj (NameMap.find var binmap)) ) then (
		let newintVal = getintVal (gettuple var obj binmap) in
		let newBitIndices = getBitIndices (gettuple var obj binmap) in
		let newBinlist = getBinlist (gettuple var obj binmap) in
		let newSpecialVar = getSpecialvar (gettuple var obj binmap) in

			(* APPEND VHDL *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = obj::var::appendvhdl in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

	var, obj, vhdl,vhdlcount,(newintVal,newBitIndices,newBinlist,newSpecialVar), binmap 

	(* if iden not found, then raise an error! *)
	) else (raise (Failure ((var)^"."^(obj)^" has no value")))  )


  | AsnRoot(var,e) -> ( 
		(* Retrive the root variable *)
		let asnvar = var in
		let asnobj = "" in


		(* APPEND PREVIOUS VHDL LINE *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = ";"::appendvhdl in
		(* Check length of of the string list! If too many operations 
			 (i.e. var := a+b+c),then raise an error!! 
 			let checklen = checktotaloperations appendvhdl in *)

	let vhdl = NameMap.add vhdlcount appendvhdl vhdl in
		(* Used for debugging 
		print_string "\n\t+ IR LINE "; print_string vhdlcount; print_string ": [";  
		print_string_list (List.rev (NameMap.find vhdlcount vhdl));

	print_string "\nAsnRoot (var,exp) -> ";*)

	(* BEGIN APPENDING A NEW VHDL LINE *)
	(* first increment the vhdl line counter *)
	let vhdlcount = increment vhdlcount in
	let appendvhdl = "<="::""::var::[] in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in


	let var, obj, vhdl,vhdlcount, tupl_rx,binmap = eval asnvar asnobj binmap vhdl vhdlcount e in


	let newbitIndices = 
		(* get the bit indices from the map if they exist *)
		if ( (NameMap.mem var binmap) && 
			 (NameMap.mem "" (NameMap.find var binmap)) ) then (
		getBitIndices(gettuple var "" binmap)
	 (* if no indices list exists, make a list *)
	) else ( mkIndices [] (List.length (getBinlist(tupl_rx))-1) ) in
	
	(* making the new tupl (slightly different than in the AsnObj procedure) *)
	let newtupl = 
	  ((getintVal(tupl_rx)),(newbitIndices),getBinlist(tupl_rx),0) in

	(* NEED TO INDICATE WHETHER OR NOT THE ROOT IS A SPECIAL VAR? *)
	(* update the binmap with the new variable and tuple *)
	let binmap = mapUpdate var "" newtupl binmap in

	(* return*)
	var, obj, vhdl, vhdlcount, newtupl, binmap)

  | AsnObj(var,obj,e) -> ( 

		(* APPEND PREVIOUS VHDL LINE *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = ";"::appendvhdl in
		(* Check length of of the string list! If too many operations 
			 (i.e. var := a+b+c),then raise an error!! 
 			let checklen = checktotaloperations appendvhdl in *)

		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in
		(* Used for debugging *)
		(*
		print_string "\n\t+ IR LINE "; print_string vhdlcount; print_string ": [";  
		print_string_list (List.rev (NameMap.find vhdlcount vhdl));*)
		
		(* print_string "\nAsnObj(var,obj,exp) -> ";*)
	(* BEGIN APPENDING A NEW VHDL LINE *)
	(* first increment the vhdl line counter *)
	let vhdlcount = increment vhdlcount in
	let appendvhdl = "<="::obj::var::[] in
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

	let var, obj, vhdl,vhdlcount, tupl_rx,binmap = eval var obj binmap vhdl vhdlcount e in

	let newbitIndices = 
		(* get the bit indices from the map *)
		if ( (NameMap.mem var binmap) && 
			 (NameMap.mem obj (NameMap.find var binmap)) ) then (
		getBitIndices(gettuple var obj binmap)
	 (* if no indices list exists, throw failure *)
	) else ( raise (Failure ((var)^"."^(obj)^" has not been declared")) )  in

	(* NOTE: NEED TO CONVERT ID INT TO BINLIST BEFORE UPDATING! 
		OTHERWISE OVERFLOW!! *)
		let newintVal = getintVal (tupl_rx) in
		let newBinlist = int2binlist newintVal in
		let newtupl = (newintVal,(newbitIndices),newBinlist,0) in

	(* update the binmap with the new variable and tuple *)
	let binmap = mapUpdate var obj newtupl binmap in


	(* return *)
	var, obj, vhdl, vhdlcount, newtupl, binmap) 
  

  | Binop(e1, op, e2) -> ( (* print_endline "\nBinop(e1,op,e2) -> "; *)

	(* Retrieve the length of the variable and object being assigned *)
	let asnvar = var in
	let asnobj = obj in
(*
	print_endline "VAR";
	print_endline var;
	print_endline obj; *)


	let var1, obj1, vhdl, vhdlcount, rxtupl1, binmap = eval var obj binmap vhdl vhdlcount e1 in

		(* Need to APPEND op to VHDL *)

	let var2, obj2, vhdl, vhdlcount, rxtupl2, binmap = 
		eval var obj binmap vhdl vhdlcount e2 in

	let newintVal = 
	  match op with
	  (* check lengths before executing the operation *)
	    Add -> ( 
			getintVal(rxtupl1) + getintVal(rxtupl2) )
	  | Mul -> ( 
			getintVal(rxtupl1) * getintVal(rxtupl2) )
	  | Sub -> (
			let checksub = 
			  if (getintVal(rxtupl1) < getintVal(rxtupl2) ) then (
				  raise (Failure ((var)^"."^(obj)^" subtraction underflow"))
			  ) else ( getintVal(rxtupl1) - getintVal(rxtupl2) ) in checksub
	  	)  
	  | Shiftr -> ( let shiftedVal = shiftR (getBinlist(rxtupl1)) (getintVal(rxtupl2)) in
				( binlist2int shiftedVal) )
	  | Shiftl -> ( let shiftedVal = shiftL (getBinlist(rxtupl1)) (getintVal(rxtupl2)) in
				( binlist2int shiftedVal)
		)   in




 	let newtupl = (newintVal,[],int2binlist(newintVal),-1) in
	  var, obj, vhdl, vhdlcount, newtupl, binmap)


  | Exprseq(e1,e2) ->( (* print_string "\nExprseq(exp1,exp2) -> "; *)
	  let var, obj, vhdl,vhdlcount, tupl_rx, binmap = eval var obj binmap vhdl vhdlcount e1 in
	  let var, obj, vhdl,vhdlcount, tupl_rx, binmap = eval var obj binmap vhdl vhdlcount e2 in
		var, obj, vhdl,vhdlcount, tupl_rx, binmap ) 
)





(* ==== Reads in the statements from the input text and handles them accordingly *)
let rec evalstatements var obj tuplrx binmap vhdl vhdlcount = (function 
	Binmap(var,objdecl) -> ( (*print_string "\n\n\n------ BINMAP(var,objdecl) ------> ";*)
	  let var,obj,indices,binmap = evalBinmap var "" [] binmap objdecl in
	  var, obj, tuplrx, vhdl, binmap )

  |	Expr(e) -> ( 
	(* print_string "\n\n\n----------- EXPRS(exp) -----------> ";*)
		let var, obj, vhdl,vhdlcount, tuplrx ,binmap = eval var obj binmap vhdl vhdlcount e in

		(* APPEND PREVIOUS VHDL LINE *)
		let appendvhdl = NameMap.find vhdlcount vhdl in
		let appendvhdl = ";"::appendvhdl in

		(* Check length of of the string list! If too many operations 
			 (i.e. var := a+b+c),then raise an error!! 
 			let checklen = checktotaloperations appendvhdl in *)
		let vhdl = NameMap.add vhdlcount appendvhdl vhdl in

		(* Used for debugging *)
		(* print_string "\n\t+ IR LINE "; print_string vhdlcount; print_string ": [";  
		print_string_list (List.rev (NameMap.find vhdlcount vhdl));
		print_endline "\n"; *)

		(* Update the binmap with the received tuple *)
	  		let rev = List.rev (appendvhdl) in
	  		let var = List.hd (rev) in
	  		let obj = List.hd (List.tl rev) in
		let binmap = mapUpdate var obj tuplrx binmap in

			var, obj, tuplrx, vhdl, binmap )

  | Stmtseq(st1,st2) ->( (* print_string "\n\nSTMTSEQ(stmt1,stmt2) -> "; *)

		let var, obj, tuplrx, vhdl, binmap = 
			evalstatements var obj tuplrx binmap vhdl vhdlcount st1 in
		let var, obj, tuplrx, vhdl, binmap = 
			evalstatements var obj tuplrx binmap vhdl vhdlcount st2 in
		var, obj, tuplrx, vhdl, binmap )

  | Print(var,obj) -> (print var obj binmap;
		var, obj, tuplrx, vhdl, binmap)
  | Printvar(var) -> (print var "" binmap;
		var, obj, tuplrx, vhdl, binmap)

  | Ifthen(condition1,condition2,ifstmt,elsestmt) -> (
		(* APPEND DUMMY VHDL? *)
	
	let var, obj, vhdl,vhdlcount, tuplrx1 ,binmap = 
		eval var obj binmap vhdl vhdlcount condition1 in
	let var, obj, vhdl,vhdlcount, tuplrx2 ,binmap = 
		eval var obj binmap vhdl vhdlcount condition2 in

	let var, obj, tuplrx, vhdl,binmap  =
		if (getintVal(tuplrx1) = getintVal(tuplrx2)) then (
			evalstatements var obj tuplrx binmap vhdl vhdlcount ifstmt
		) else (evalstatements var obj tuplrx binmap vhdl vhdlcount elsestmt) in
		
		var, obj, tuplrx, vhdl, binmap)

  | Repeat(repeatstmt,condition1,condition2) -> (

	let var, obj, vhdl,vhdlcount, tuplrx1 ,binmap = 
		eval var obj binmap vhdl vhdlcount condition1 in
	let var, obj, vhdl,vhdlcount, tuplrx2 ,binmap = 
		eval var obj binmap vhdl vhdlcount condition2 in

	let cond1 = getintVal(tuplrx1) in
	let cond2 = getintVal(tuplrx2) in

	

	let binmap = 
	  let rec loop binmap var obj vhdl vhdlcount repeatstmt cond1 cond2 out = 


			if (cond1 = cond2) then (
				binmap
			) else (

				let var, obj, tuplrx, vhdl, binmap = 
					evalstatements var obj tuplrx binmap vhdl vhdlcount repeatstmt in

			  	let _, _, _, _, tuplrx2, _ = 
					eval var obj binmap vhdl vhdlcount condition1 in
			
				let _, _, _, _, tuplrx1, _ = 
					eval var obj binmap vhdl vhdlcount condition2 in

				let cond1 = getintVal(tuplrx1) in
				let cond2 = getintVal(tuplrx2) in
		
				(* let intUpdate = getintVal (gettuple var obj binmap) 
				let intUpdate = getintVal (tuplrx) in

				print_string "\nVAL var:\t";
				print_int (getintVal ( gettuple "var" "" binmap ));

				print_string "\nVAL var2:\t";
				print_int (intUpdate);

				(* check binary map *)
				let binmap = mapUpdate var obj tuplrx binmap in*)
	
			loop binmap var obj vhdl (increment vhdlcount) repeatstmt cond1 cond2 out ) in 
		loop binmap var obj vhdl vhdlcount repeatstmt cond1 cond2 [] in 

	var, obj, tuplrx, vhdl, binmap

	)

)




let _ =
  let lexbuf = Lexing.from_channel stdin in
  let statement = Parser.statement Scanner.token lexbuf in

  let var,obj,tuplrx, vhdl, binmap = 
	evalstatements "" "" (0,[],[],0) binmap vhdl vhdlcount statement in
	let var = "var1" in
	let obj = "" in []




